from __future__ import print_function
"""
Tubing sources are defined here. If you want to make your own Source, create a
reader class with a read(amt) function, where amt is the amount of `stuff` to
read. MakeSource can generate a Source from your Reader. Ex::

    class MyReader(object):
        def read(self, amt):
            return [0] * amt, False

    MySource = MakeSource(MyReader)

The read(amt) function should return a chunk of data and a boolean indicated if
we've reached EOF. Unlike normal python streams, it's ok to return empty sets
and it won't close the stream. Only returning True as the second parameter
indicates that the stream is closed.
"""
import logging

logger = logging.getLogger('tubing.sources')


class MakeSource(object):
    """
    MakeSource takes a reader object and returns a Source factory.
    """

    def __init__(self, reader_cls):
        self.reader_cls = reader_cls

    def __call__(self, *args, **kwargs):
        return Source(self.reader_cls(*args, **kwargs))


class Source(object):
    """
    Source is a wrapper for Readers that allows piping.
    """

    def __init__(self, source):
        self.source = source

    def read(self, amt):
        return self.source.read(amt)

    def __or__(self, other):
        return self.pipe(other)

    def pipe(self, other):
        return other(self)


class ObjectReader(object):
    """
    ObjectReader outputs a list of objects.
    """

    def __init__(self, objs):
        self.objs = objs

    def read(self, amt=None):
        r, self.objs = self.objs[:amt], self.objs[amt or len(self.objs):]
        return r, len(self.objs) == 0


Objects = MakeSource(ObjectReader)



from __future__ import print_function
"""
Tubing sinks are targets for streams of data. To make your own sink, define a Writer with a write(chunk) and optional close() and abort() functions and pass it to MakeSink. if present, close() will be called on the upstream EOF and abort() will be called on any exception in the pipeline. Here's an example::

    class MyWriter(object):
        def write(self, chunk):
            print chunk

    MySink = MakeSink(MyWriter)
"""
import logging
import io
import functools
import requests

logger = logging.getLogger('tubing.sinks')


class MakeSink(object):
    """
    MakeSink returns a factory that wraps a Writer and executes the pipeline.
    default_chunk_size should be set according to the type of stream that the
    sink works on. Objects are heavier than byte streams, so should have a
    lower default_chunk_size.
    """

    def __init__(self, sink_cls, default_chunk_size=2**16):
        self.sink_cls = sink_cls
        self.default_chunk_size = default_chunk_size

    def __call__(self, *args, **kwargs):
        if kwargs.get('chunk_size'):
            chunk_size = kwargs['chunk_size']
            del kwargs['chunk_size']
        else:
            chunk_size = None

        chunk_size = chunk_size or self.default_chunk_size

        def fn(source):
            try:
                sink = self.sink_cls(*args, **kwargs)
                logger.debug("reading %s" % (source))
                chunk, eof = source.read(chunk_size)
                sink.write(chunk)
                while not eof:
                    chunk, eof = source.read(chunk_size)
                    sink.write(chunk)
                hasattr(sink, 'close') and sink.close()
                return sink
            except:
                logger.exception("Pipe failed")
                hasattr(sink, 'abort') and sink.abort()
                raise

        return fn


def GeneratorSink(chunk_size, source):
    eof = False
    while not eof:
        r, eof = source.read(chunk_size)
        yield r


def Generator(chunk_size=2 ** 10):
    return functools.partial(GeneratorSink, chunk_size)


def GeneratorGeneratorSink(chunk_size, per_gen, flatten, source):
    d = dict(eof=False)
    def gen():
        i = per_gen
        while not d['eof'] and i:
            chunk, d['eof'] = source.read(chunk_size)
            logger.debug(repr(chunk))
            if flatten:
                for line in chunk:
                    yield line
            else:
                yield chunk
            i -= 1

    while not d['eof']:
        yield gen()

    logger.debug("all done")

def GeneratorGenerator(chunk_size=2 ** 10, per_gen=2 * 4, flatten=False):
    return functools.partial(GeneratorGeneratorSink, chunk_size, per_gen, flatten)


class ObjectsSink(list):
    """
    Writes an object stream to a list.
    """

    def write(self, objs):
        self.extend(objs)


Objects = MakeSink(ObjectsSink, 2**4)


class BytesWriter(io.BytesIO):
    """
    BytesWriter collects a stream of bytes and saves it to the result member
    on close().
    """

    def __init__(self, *args, **kwargs):
        self.result = None
        super(BytesWriter, self).__init__(*args, **kwargs)

    def close(self):
        self.result = self.getvalue()
        super(BytesWriter, self).close()

    def abort(self):
        super(BytesWriter, self).close()


Bytes = MakeSink(BytesWriter, 2**16)


class FileWriter(object):
    def __init__(self, *args, **kwargs):
        self.f = open(*args, **kwargs)

    def write(self, chunk):
        self.f.write(chunk)

    def close(self):
        self.f.close()

    def abort(self):
        # TODO: probably delete the file
        self.f.close()


File = MakeSink(FileWriter)


class DebugPrinter(object):
    def write(self, chunk):
        logger.debug(chunk)

    def close(self):
        logger.debug("CLOSED")

    def abort(self):
        logger.debug("ABORTED")


Debugger = MakeSink(DebugPrinter)


class HTTPPost(object):
    """
    Expects a stream of byte strings.
    """
    def __init__(self, url, username=None, password=None, chunk_size=2 ** 4, chunks_per_post=2 ** 10):
        self.url = url
        self.auth = None
        if username and password:
            self.auth = username, password
        self.chunk_size = chunk_size
        self.per_post = chunks_per_post

    def __call__(self, source):
        posts = GeneratorGeneratorSink(
            chunk_size=self.chunk_size,
            per_gen=self.per_post,
            flatten=True,
            source=source
        )
        for post in posts:
            resp = requests.post(self.url, data=post, auth=self.auth)



from __future__ import print_function
"""
S3 Tubing extension.
"""

import boto3
import logging
from tubing import sources, sinks

logger = logging.getLogger('tubing.ext.s3')


class S3Reader(object):  # pragma: no cover
    """
    Read file from S3. Expects AWS environmental variables to be set.
    """

    def __init__(self, bucket, key):  # pragma: no cover
        """
        Create an S3 Source stream.
        """
        s3 = boto3.client('s3')
        self.response = s3.get_object(Bucket=bucket, Key=key)

    def read(self, amt):
        r = self.response['Body'].read(amt)
        return r or b'', not r


S3Source = sources.MakeSource(S3Reader)


class S3SinkWriter(object):  # pragma: no cover
    """
    Send file to S3. Expects AWS environmental variables to be set.
    """

    def __init__(self, bucket, key):
        """
        Initiate the multipart upload.
        """
        self.s3 = boto3.client('s3')
        self.bucket = bucket
        self.key = key
        logger.debug("Creating upload for %s %s" % (bucket, key))
        upload = self.s3.create_multipart_upload(Bucket=bucket, Key=key)
        self.upload_id = upload["UploadId"]
        # start with 1
        self.part_number = 1
        self.part_info = dict(Parts=[])

    def track_part(self, part_number, part):
        """
        Keep track of all parts so we can finalize multipart upload.
        """
        self.part_info['Parts'].append(
            dict(
                PartNumber=part_number,
                ETag=part['ETag'],
            )
        )

    def write(self, chunk):
        """
        Upload a part.
        """
        logger.debug("Posting " + self.part_number)
        part = self.s3.upload_part(
            Bucket=self.bucket,
            Key=self.key,
            PartNumber=self.part_number,
            UploadId=self.upload_id,
            Body=chunk,
        )
        self.track_part(self.part_number, part)
        self.part_number += 1

    def close(self):
        """
        Finalize upload.
        """
        self.s3.complete_multipart_upload(
            Bucket=self.bucket,
            Key=self.key,
            UploadId=self.upload_id,
            MultipartUpload=self.part_info,
        )

    def abort(self):
        """
        Something failed, abort!
        """
        self.s3.abort_multipart_upload(
            Bucket=self.bucket,
            Key=self.key,
            UploadId=self.upload_id,
        )


S3Sink = sinks.MakeSink(S3SinkWriter)
